from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Tuple


class ABCCommandParser(ABC):
    @abstractmethod
    def __call__(self, raw_command: str) -> List[str]:
        pass


class ABCCommandArgumentsCaster(ABC):
    @abstractmethod
    def __call__(self, fn: Callable, tokens: List[str]) -> Tuple[Tuple[Any], Dict[str, Any]]:
        pass


class ABCCommandHandler(ABC):
    @abstractmethod
    def handle(self, fn: Callable[..., Any], text: str):
        pass



from typing import TYPE_CHECKING, List

from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.completion.base import CompleteEvent
from prompt_toolkit.document import Document

from eggella.exceptions import CommandNotFoundError

if TYPE_CHECKING:
    from eggella.manager import CommandManager


class CommandCompleter(Completer):
    def __init__(self, manager: "CommandManager", ignore_case: bool = True):
        self.manager = manager
        self.ignore_case = ignore_case

    def get_completions(self, document: Document, complete_event: CompleteEvent):
        text_before_cursor = document.text_before_cursor
        if self.ignore_case:
            text_before_cursor = text_before_cursor.lower()
        text_arr = text_before_cursor.split(" ")
        last_words = text_arr[-1]
        completions = self.__get_current_completions(text_arr[:-1])

        for completion, meta in completions:
            if completion not in document.text_before_cursor and "=" not in last_words:
                yield Completion(completion, -len(last_words), display_meta=meta or "")

    def __get_current_completions(self, text_arr: List[str]):
        if not text_arr:
            return self.manager.all_completions
        command = text_arr[0]
        try:
            command_obj = self.manager.get(command)
        except CommandNotFoundError:
            return []
        if command_obj:
            return [command_obj.completion]



from typing import Any, Callable, Dict, Optional, Tuple

from eggella.command.abc import (
    ABCCommandArgumentsCaster,
    ABCCommandHandler,
    ABCCommandParser,
)
from eggella.command.command import CommandArgumentsCaster, CommandParser


class CommandHandler(ABCCommandHandler):
    def __init__(
        self,
        parser: ABCCommandParser = CommandParser(),
        caster: Optional[ABCCommandArgumentsCaster] = CommandArgumentsCaster(),
    ):
        self.parser = parser
        self.caster = caster

    def handle(self, fn: Callable[..., Any], text: str) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        arguments = self.parser(text)
        return self.caster(fn, arguments) if self.caster else (tuple(arguments), {})



from difflib import SequenceMatcher
from typing import Any, Iterable

from prompt_toolkit import print_formatted_text as print_ft
from prompt_toolkit.formatted_text.html import HTML
from prompt_toolkit.shortcuts.prompt import confirm
from prompt_toolkit.styles import Style

from eggella.events.abc import ABCEvent


class OnStartup(ABCEvent):
    _STYLES = Style.from_dict({"help": "#80ff00", "close": "#ff0000", "cmd_key": "#7d7c80 italic"})

    def __call__(self):
        pass


class OnClose(ABCEvent):
    def __call__(self):
        pass


class OnKeyboardInterrupt(ABCEvent):
    def __call__(self) -> bool:
        try:
            if confirm("Exit?"):
                return True
        except (KeyboardInterrupt, EOFError):
            return True
        return False


class OnEOFError(ABCEvent):
    def __call__(self):
        try:
            if confirm("Exit?"):
                exit(0)
        except (KeyboardInterrupt, EOFError):
            exit(0)


class OnCommandError(ABCEvent):
    def __call__(self, key: str, inline_arguments: str):
        pass


class OnCommandNotFound(ABCEvent):
    _STYLE = Style.from_dict(
        {
            "cmd_key": "#7d7c80 italic",
        }
    )

    def __call__(self, key: str, inline_arguments: str):
        print_ft(
            HTML(f"<ansired>Error!</ansired> command <cmd_key>{key}</cmd_key> not founded"),
            style=self._STYLE,
        )


class OnSuggest(ABCEvent):
    _STYLE = Style.from_dict({"mean": "#ffff00 bold"})

    def __call__(self, command: str, possible_commands: Iterable[str]):
        suggested_command = None
        ratio = 0
        for possible_command in possible_commands:
            possible_ratio = SequenceMatcher(None, command, possible_command).ratio()
            if possible_ratio > ratio:
                ratio = possible_ratio  # type: ignore
                suggested_command = possible_command
        print_ft(
            HTML(f"Did your mean: <mean>{suggested_command}</mean> ?"),
            style=self._STYLE,
        )


class OnCommandCompleteSuccess(ABCEvent):
    _STYLE = Style.from_dict({"out": "#696969 bold"})

    def __call__(self, result: Any):
        if result:
            print_ft(HTML(f"<out>{result}</out>"), style=self._STYLE)


class OnCommandArgumentsException(ABCEvent):
    def __call__(self, key: str, error: BaseException, *args, **kwargs):
        pass



import shlex
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
)

from eggella.command.abc import ABCCommandHandler
from eggella.command.completer import CommandCompleter
from eggella.command.handler import CommandHandler
from eggella.command.objects import Command
from eggella.events.events import (
    OnCommandCompleteSuccess,
    OnCommandError,
    OnCommandNotFound,
    OnEOFError,
    OnKeyboardInterrupt,
    OnSuggest,
)
from eggella.exceptions import CommandNotFoundError
from eggella.shortcuts.help_pager import gen_help_pager

if TYPE_CHECKING:
    from eggella.app import Eggella


_ErrorEventsMapping = Dict[str, Tuple[Type[BaseException], ...]]


class CommandManager:
    def __init__(self, app: "Eggella"):
        self._app = app
        self.commands: Dict[str, Command] = {}
        self.error_events: Dict[str, Callable[[str, BaseException, str, str], Any]] = {}
        self.handled_exceptions: _ErrorEventsMapping = {}
        self._register_buildin_commands()

    @staticmethod
    def _simple_parse_arguments(raw_command: str) -> Tuple[Tuple[str, ...], Dict[str, str]]:
        tokens = shlex.split(raw_command)
        args: List[str] = []
        kwargs: Dict[str, str] = {}
        for token in tokens:
            if "=" in token:
                key, value = token.split("=", 1)
                kwargs[key] = value
            else:
                args.append(token)
        return tuple(args), kwargs

    def exec(self, key: str, args: str):
        command = self.get(key)
        try:
            return command.handle(args)
        except BaseException as e:
            if err_handler := self.error_events.get(command.fn.__name__):
                handle_exceptions = self.handled_exceptions.get(command.fn.__name__, None)
                if handle_exceptions and any(e.__class__ == exc for exc in handle_exceptions):
                    _args, _kwargs = self._simple_parse_arguments(args)
                    return err_handler(key, e, *_args, **_kwargs)
                else:
                    raise e
            raise e

    def get_completer(self) -> CommandCompleter:
        return CommandCompleter(self)  # type: ignore

    def on_error(self, *errors: Type[BaseException]):
        def decorator(handler: Callable[[str, BaseException, str, str], Any]):
            @wraps(handler)
            def decorator_wrapper(func: Callable[..., Any]):
                if not self.error_events.get(func.__name__):
                    self.error_events[func.__name__] = handler
                    self.handled_exceptions[func.__name__] = errors

                @wraps(func)
                def wrapper(*args, **kwargs):
                    try:
                        return func(*args, **kwargs)
                    except BaseException as e:
                        if any(e.__class__ == exc for exc in errors):
                            return handler("", e, *args, **kwargs)
                        raise e

                return wrapper

            return decorator_wrapper

        return decorator

    @property
    def all_completions(self) -> List[Tuple[str, str]]:
        return [com.completion for com in self.commands.values()]

    def get(self, key: str) -> Command:
        if command := self.commands.get(key, None):
            return command
        raise CommandNotFoundError(f"Command {key} not founded")

    def command(
        self,
        key: Optional[str] = None,
        *,
        short_description: Optional[str] = None,
        usage: Optional[str] = None,
        cmd_handler: Optional[ABCCommandHandler] = None,
    ):
        def decorator(func: Callable):
            self.register_command(
                func,
                key=key,
                short_description=short_description,
                usage=usage,
                cmd_handler=cmd_handler,
            )

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def register_command(
        self,
        func: Callable,
        key: Optional[str] = None,
        *,
        short_description: Optional[str] = None,
        usage: Optional[str] = None,
        cmd_handler: Optional[ABCCommandHandler] = None,
    ):
        if not key:
            key = func.__name__
        if self.commands.get(key):
            raise TypeError(f"Command '{key}' already register")

        if not cmd_handler:
            self.commands[key] = Command(
                fn=func,
                key=key,
                handler=CommandHandler(),
                short_description=short_description,
                usage=usage,
            )
        elif cmd_handler:
            self.commands[key] = Command(
                fn=func,
                key=key,
                handler=cmd_handler,
                short_description=short_description,
                usage=usage,
            )

    def _help_command(self, key: Optional[str] = None):
        """show help or show pager documentation for all commands if not argument passed"""
        if not key:
            commands = self.commands.values()
            gen_help_pager(self._app, commands)
            return
        elif comma := self.commands.get(key):
            return f"{key} - {comma.short_desc}"
        else:
            raise CommandNotFoundError

    @staticmethod
    def _exit_command():
        """exit from this application"""
        raise KeyboardInterrupt

    def _register_buildin_commands(self):
        self.register_command(self._help_command, "help", usage="help; help exit")
        self.register_command(self._exit_command, "exit")


class EventManager:
    def __init__(self, app: "Eggella"):
        self.app = app
        self.startup_events: List[Callable] = []
        self.close_events: List[Callable] = []
        self.errors_events: Dict[str, Callable] = {}

        self.kb_interrupt_event: Callable[..., bool] = OnKeyboardInterrupt()
        self.eof_event: Callable[..., bool] = OnEOFError()
        self.command_error_event: Callable[..., None] = OnCommandError()
        self.command_not_found_event: Callable[..., None] = OnCommandNotFound()
        self.command_complete_event: Callable[..., None] = OnCommandCompleteSuccess()
        self.command_suggest_event: Optional[Callable[..., None]] = OnSuggest()

    def register_event(
        self,
        name: Literal[
            "start", "close", "kb_interrupt", "eof", "command_not_found", "command_complete", "command_suggest"
        ],
        func: Optional[Callable],
    ):
        if name == "start" and func:
            self.startup_events.append(func)
            return
        elif name == "close" and func:
            self.close_events.append(func)
            return
        elif name == "kb_interrupt" and func:
            self.kb_interrupt_event = func
            return
        elif name == "eof" and func:
            self.eof_event = func
            return
        elif name == "command_not_found" and func:
            self.command_not_found_event = func
            return
        elif name == "command_complete" and func:
            self.command_complete_event = func
            return
        elif name == "command_suggest":
            self.command_suggest_event = func
            return
        raise TypeError("Unknown event type")

    def startup(self):
        def decorator(func: Callable):
            self.register_event("start", func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def close(self):
        def decorator(func: Callable):
            self.register_event("close", func)

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return decorator



from typing import TYPE_CHECKING, Iterable

# https://github.com/prompt-toolkit/python-prompt-toolkit/blob/master/examples/full-screen/pager.py
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import SearchToolbar, TextArea

if TYPE_CHECKING:
    from eggella import Eggella
    from eggella.command.objects import Command


def _render_text(app: "Eggella", commands: Iterable["Command"]):
    """
    {{APP DOCUMENTATION}}

    COMMANDS
       {{comma1}} - [arg1] [arg2] - {{short description}}
           {{description}}
           {{description}}
           {{description}}
       Usage:
           {{USAGE}}
           {{USAGE}}

       {{comma2}} - [arg1] [arg2] - {{short description}}
           {{description}}
           {{description}}
           {{description}}
       Usage:
           {{USAGE}}
           {{USAGE}}

       {{comma3}} - [arg1] [arg2] - {{short description}}
           {{description}}
           {{description}}
           {{description}}

    """

    text = app.documentation
    text += "\n"
    text += "COMMANDS:\n"
    for command in commands:
        args = " ".join([f"[{arg}]" for arg in command.arguments])
        text += f"    {command.key} {args}\n"
        if command.short_description:
            for line in command.short_description.split("\n"):
                text += f"        {line}\n"
        else:
            for line in command.docstring.split("\n"):
                text += f"        {line}\n"
        text += "\n"
        if command.usage:
            text += "        USAGE:\n"
            for line in command.usage.split("\n"):
                text += f"            {line}\n"
            text += "\n"
    return text


def gen_help_pager(app: "Eggella", commands: Iterable["Command"]):
    text = _render_text(app, commands)
    search_field = SearchToolbar(text_if_not_searching=[("class:not-searching", "Press '/' to start searching.")])

    text_area = TextArea(
        text=text,
        read_only=True,
        scrollbar=True,
        search_field=search_field,
    )
    status_bar_text = [
        ("class:status", "Help page"),
        ("class:status", " - Press "),
        ("class:status.key", "Ctrl-C or Q"),
        ("class:status", " to exit, "),
        ("class:status.key", "/"),
        ("class:status", " for searching."),
    ]
    root_container = HSplit(
        [
            # The top toolbar.
            Window(
                content=FormattedTextControl(status_bar_text),  # type: ignore
                height=D.exact(1),
                style="class:status",
            ),
            # The main content.
            text_area,
            search_field,
        ]
    )

    # Key bindings.
    bindings = KeyBindings()

    @bindings.add("c-c")
    @bindings.add("q")
    def _(event):
        "Quit."
        event.app.exit()

    style = Style.from_dict(
        {
            "status": "reverse",
            "status.position": "#aaaa00",
            "status.key": "#ffaa00",
            "not-searching": "#888888",
        }
    )

    # create application.
    application = Application(  # type: ignore
        layout=Layout(root_container, focused_element=text_area),
        key_bindings=bindings,
        enable_page_navigation_bindings=True,  # type: ignore
        mouse_support=True,
        style=style,
        full_screen=True,
    )
    application.run()



from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    overload,
)

from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.completion.fuzzy_completer import FuzzyCompleter
from prompt_toolkit.formatted_text import FormattedText

from eggella.command.abc import ABCCommandHandler
from eggella.exceptions import CommandNotFoundError, CommandParseError
from eggella.fsm.fsm import FsmController, IntState
from eggella.manager import CommandManager, EventManager
from eggella.shortcuts.cmd_shortcuts import CmdShortCuts

PromptLikeMsg = Union[str, FormattedText, Callable[..., Union[FormattedText, List[Tuple[str, str]]]]]

_DEFAULT_INTRO_MSG = HTML(
    "<ansired>Press |CTRL+C| or |CTRL+D| or type</ansired> exit <ansired>for close this app</ansired>\n"
    "<ansigreen>Press |TAB| or type</ansigreen> help <ansigreen>for get commands information</ansigreen>"
)


class Eggella:
    __app_instances__: Dict[str, "Eggella"] = {}

    def __new__(cls, app_name: str, msg: PromptLikeMsg = "~> ") -> "Eggella":
        if _app := cls.__app_instances__.get(app_name):
            return _app

        new_app = super().__new__(cls)
        cls.__app_instances__[app_name] = new_app
        return new_app

    def __init__(self, app_name: str, msg: PromptLikeMsg = "~> "):
        self.app_name = app_name
        self.prompt_msg = msg
        self.session: PromptSession = PromptSession(msg)
        self.cmd = CmdShortCuts()

        self.CTX: Dict[Hashable, Any] = {}
        self._intro: Union[HTML, PromptLikeMsg] = _DEFAULT_INTRO_MSG
        self._doc: str = ""
        # managers
        self._command_manager: CommandManager = CommandManager(self)
        self._event_manager = EventManager(self)

        # fsm
        self.fsm = FsmController(self)

    @property
    def documentation(self):
        return self._doc

    @documentation.setter
    def documentation(self, text: str):
        self._doc = text

    @property
    def intro(self):
        return self._intro

    @intro.setter
    def intro(self, text: Union[HTML, PromptLikeMsg]):
        self._intro = text

    def on_startup(self):
        return self._event_manager.startup()

    def on_close(self):
        return self._event_manager.close()

    def on_error(self, *errors: Type[BaseException]):
        return self._command_manager.on_error(*errors)

    def on_command(
        self,
        key: Optional[str] = None,
        short_description: Optional[str] = None,
        *,
        usage: Optional[str] = None,
        cmd_handler: Optional[ABCCommandHandler] = None,
    ):
        return self._command_manager.command(
            key,
            short_description=short_description,
            usage=usage,
            cmd_handler=cmd_handler,
        )

    def on_state(self, state: IntState):
        return self.fsm.state(state)

    def register_states(self, states: Type[IntState]):
        self.fsm.attach(states)

    def register_command(
        self,
        func: Callable,
        key: Optional[str] = None,
        short_description: Optional[str] = None,
        *,
        usage: Optional[str] = None,
        cmd_handler: Optional[ABCCommandHandler] = None,
    ):
        self._command_manager.register_command(
            func,
            key,
            short_description=short_description,
            usage=usage,
            cmd_handler=cmd_handler,
        )

    @overload
    def register_event(
        self,
        name: Literal["start", "close", "kb_interrupt", "eof", "command_not_found", "command_complete"],
        func: Callable,
    ) -> None:
        pass

    @overload
    def register_event(
        self,
        name: Literal["command_suggest"],
        func: Optional[Callable],
    ) -> None:
        pass

    def register_event(
        self,
        name: Literal[
            "start", "close", "kb_interrupt", "eof", "command_not_found", "command_complete", "command_suggest"
        ],
        func: Optional[Callable],
    ) -> None:
        self._event_manager.register_event(name, func)

    def has_command(self, key: str) -> bool:
        return bool(self._command_manager.commands.get(key, None))

    def remove_command(self, key: str):
        if self.has_command(key):
            self._command_manager.commands.pop(key)
        else:
            raise KeyError

    def loop(self):
        self.cmd.print_ft(self.intro)
        self._handle_startup_events()
        self._handle_commands()
        self._handle_close_events()

    def _handle_startup_events(self):
        for event in self._event_manager.startup_events:
            event()

    def _handle_close_events(self):
        for event in self._event_manager.close_events:
            event()

    def _handle_commands(self):
        while True:
            try:
                completer = FuzzyCompleter(completer=self._command_manager.get_completer())
                result = self.session.prompt(self.prompt_msg, completer=completer)
                if not result:
                    continue

                if (tokens := result.split(" ", 1)) and len(tokens) == 1:
                    key = tokens[0]
                    args = ""
                else:
                    key, args = tokens[0], tokens[1]

                if result := self._command_manager.exec(key, args):
                    self._event_manager.command_complete_event(result)
            except CommandNotFoundError:
                self._event_manager.command_not_found_event(key, args)
                self._event_manager.command_suggest_event(key, self._command_manager.commands.keys())
            except CommandParseError:
                self._event_manager.command_error_event(key, args)
            except KeyboardInterrupt:
                if self._event_manager.kb_interrupt_event():
                    break
            except EOFError:
                self._event_manager.eof_event()



from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, Union

if TYPE_CHECKING:
    from eggella.app import Eggella


class IntState(int, Enum):
    @classmethod
    def last(cls) -> "IntState":
        return cls.list()[-1]

    @classmethod
    def list(cls) -> List["IntState"]:
        return list(cls.__iter__())

    @classmethod
    def first(cls) -> "IntState":
        return cls.list()[0]


class Fsm:
    def __init__(self, state: Type["IntState"]):
        self.ctx: Dict[str, Any] = {}
        self.handlers: Dict[IntState, Callable] = {}

        self._all_states = state.list()
        self._start_state = state.first()
        self._end_state = state.last()
        self._current_state: Optional[IntState] = None

    def add_handler(self, state: IntState, func: Callable):
        self.handlers[state] = func

    def is_finish(self) -> bool:
        return self._current_state is None

    def on_state(self, state: "IntState"):
        def decorator(func):
            self.handlers[state] = func
            return func

        return decorator

    def set(self, state: "IntState"):
        self._current_state = state
        return self.handlers[self._current_state]()

    def clear(self):
        self.ctx.clear()

    def run(self, state: Optional["IntState"] = None):
        self._current_state = state or self._start_state
        return self.handlers[self._current_state]()

    def finish(self) -> None:
        self.ctx.clear()
        self._current_state = None

    def next(self):
        index = self._all_states.index(self._current_state)
        if index < len(self._all_states):
            self._current_state = self._all_states[index + 1]
            return self.handlers[self._current_state]()
        else:
            raise IndexError("States out of range")

    def prev(self):
        index = self._all_states.index(self._current_state)
        if index < len(self._all_states) and index != 0:
            self._current_state = self._all_states[index - 1]
            return self.handlers[self._current_state]()
        else:
            raise IndexError("States out of range")

    def __getitem__(self, item):
        return self.ctx[item]

    def __setitem__(self, key, value):
        self.ctx[key] = value


class FsmController:
    def __init__(self, app: "Eggella"):
        self.fsm_storage: Dict[str, Fsm] = {}
        self._current_fsm: Optional[Fsm] = None
        self.__app = app

    def __getitem__(self, item):
        if self._current_fsm:
            return self._current_fsm[item]
        raise AttributeError("Need activate FSM first")

    def __setitem__(self, key, value):
        if not self._current_fsm:
            raise AttributeError("Need invoke FSM first")
        self._current_fsm[key] = value

    def is_active(self):
        return self._current_fsm is not None

    @property
    def ctx(self):
        if self._current_fsm:
            return self._current_fsm.ctx
        raise AttributeError("Need invoke FSM first")

    def attach(self, states: Type[IntState]):
        self.fsm_storage[states.__name__] = Fsm(states)

    def state(self, state: IntState):
        """state decorator"""
        return self.fsm_storage[state.__class__.__name__].on_state(state)

    def run(self, state: Union[IntState, Type[IntState]]):
        if isinstance(state, IntState):
            self._current_fsm = self.fsm_storage[state.__class__.__name__]
            return self._current_fsm.run(state)
        else:
            self._current_fsm = self.fsm_storage[state.__name__]
            return self._current_fsm.run(state.first())

    def prev(self):
        if not self._current_fsm:
            raise AttributeError("Need activate FSM first")
        self._current_fsm.prev()

    def next(self):
        if not self._current_fsm:
            raise AttributeError("Need activate FSM first")
        self._current_fsm.next()

    def set(self, state: IntState):
        if not self._current_fsm:
            raise AttributeError("Need activate FSM first")
        self._current_fsm.set(state)

    def finish(self):
        if not self._current_fsm:
            raise AttributeError("Need activate FSM first")
        self._current_fsm.finish()
        self._current_fsm = None


if __name__ == "__main__":

    class MyStates(IntState):
        a = 1
        b = 2

    print(type(MyStates))
    print(type(MyStates.a))
    # print(MyStates.__name__)
    # print(MyStates.first().__class__.__name__)
    # print(MyStates.a.__class__.__name__)



from eggella import Eggella

app = Eggella(__name__)
app.documentation = """Welcome to simple application
    this app does nothing

"""


@app.on_startup()
def startup_event():
    print("Hello! this quick Eggella app example!")
    print("Press Ctrl+C or Ctrl+D or type `exit` for close application")


@app.on_close()
def close_event():
    print("Goodbye! :D")


@app.on_command()
def hello():
    return "Hello, world!"


@app.on_error(ValueError, TypeError)
# first argument - command key, second - handled exception
def my_error_handler(key: str, exc: BaseException, *args: str, **kwargs: str):
    if key == "sum":
        bad_args = [arg for arg in args if not str(arg).isdigit()]
        print("Error! sum accept bad arguments:", *bad_args)
        return
    print(exc)


@app.on_error(ConnectionError)
def dummy_handle_exc(key: str, exc: BaseException, *args, **kwargs):
    print("catch connection error")
    return ":P"


@my_error_handler
@app.on_command("sum")
def sum_(*args: int):
    """sum all passed digits. if argument is not str - print error message"""
    return sum(args)


@my_error_handler
@app.on_command("raise-err1")
def raise_err1():
    """raise TypeError and catch exception"""
    raise TypeError("TypeError - my_error_handler catch this")


@dummy_handle_exc
@app.on_command("raise-err2")
def raise_err2():
    """raise ConnectionError and catch exception"""

    raise ConnectionError("connect error - dummy_handle_exc catch this")


@dummy_handle_exc  # decorators stack not work
@my_error_handler
@app.on_command("raise-err3")
def raise_err3():
    """raise TypeError and close app"""
    raise ConnectionError("connect error - dummy_handle_exc not catch this :(")


if __name__ == '__main__':
    app.loop()



import ast
import inspect
import shlex
from contextlib import suppress
from typing import Any, Callable, Dict, List, Tuple

from eggella.command.abc import ABCCommandArgumentsCaster, ABCCommandParser
from eggella.tools.type_caster import TypeCaster


class CommandParser(ABCCommandParser):
    def __call__(self, raw_command: str) -> List[str]:
        return shlex.split(raw_command)


class CommandParserRaw(ABCCommandParser):
    def __call__(self, raw_command: str) -> List[str]:
        return [raw_command]


class CommandArgumentsCaster(ABCCommandArgumentsCaster):
    @staticmethod
    def _literal_eval(token: str) -> Any:
        # simple convert variables like list, dict, int, float
        with suppress(ValueError, SyntaxError):
            token = ast.literal_eval(token)
            return ast.literal_eval(token)
        return token

    @staticmethod
    def _cast_arguments(fn, *args, **kwargs):
        sig = inspect.signature(fn)
        params = sig.parameters
        tc = TypeCaster()
        bound = sig.bind(*args, **kwargs)
        # set defaults, if not set
        bound.apply_defaults()
        for arg_name, values in bound.arguments.items():
            if param := params.get(arg_name):
                # *args
                if isinstance(values, tuple) and param.kind is param.VAR_POSITIONAL:
                    bound.arguments[arg_name] = tuple(tc.cast(param.annotation, v) for v in values)
                # **kwargs
                elif isinstance(values, dict) and param.kind is param.VAR_KEYWORD:
                    bound.arguments[arg_name] = {k: tc.cast(param.annotation, v) for k, v in values.items()}
                # positional
                else:
                    bound.arguments[arg_name] = tc.cast(param.annotation, values)
        return bound.args, bound.kwargs

    def __call__(self, fn: Callable, tokens: List[str]) -> Tuple[Tuple[Any], Dict[str, Any]]:
        sig = inspect.signature(fn)
        param_names = list(sig.parameters.keys())
        args: List[Any] = []
        kwargs = {}

        for token in tokens:
            if "=" in token:
                key, value = token.split("=", 1)
                kwargs[key] = self._literal_eval(value)
            elif len(args) < len(param_names) and param_names[len(args)] not in kwargs:
                args.append(self._literal_eval(token))
            else:
                args.append(self._literal_eval(token))
        # bind parsed arguments, set defaults values
        bound = sig.bind(*args, **kwargs)
        # set defaults, if not set
        bound.apply_defaults()
        return self._cast_arguments(fn, *bound.args, **bound.kwargs)


if __name__ == "__main__":
    cp = CommandParser()
    print(cp("a='12 123' 100"))



import inspect
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple

from eggella.command.abc import ABCCommandHandler
from eggella.command.handler import CommandHandler


class Command(NamedTuple):
    fn: Callable[..., Any]
    key: str
    handler: ABCCommandHandler = CommandHandler()
    usage: Optional[str] = None
    short_description: Optional[str] = None

    def handle(self, command_text: str) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        args, kwargs = self.handler.handle(self.fn, command_text)
        return self.fn(*args, **kwargs)

    @property
    def arguments(self) -> List[str]:
        args = inspect.signature(self.fn).parameters.values()
        return [str(arg) for arg in args]

    @property
    def docstring(self) -> str:
        return inspect.getdoc(self.fn) or ""

    @property
    def short_desc(self):
        arg_list = " ".join(f"[{arg}]" for arg in self.arguments)
        if self.short_description:
            return f"({arg_list}) - {self.short_description}"
        elif self.docstring:
            short_desc = self.docstring.split("\n")[0]
            return f"{arg_list} - {short_desc}"
        return f"{arg_list}"

    @property
    def completion(self) -> Tuple[str, str]:
        return self.key, self.short_desc

    @property
    def help(self):
        arg_list = " ".join(f"[{arg}]" for arg in self.arguments)
        if len(self.docstring.split("\n")) > 1:
            if self.usage:
                return f"      {self.key} ({arg_list})\n            {self.docstring}\nUSAGE:\n      {self.usage}"
            return f"      {self.key} ({arg_list})\n            {self.docstring}"
        if self.usage:
            return f"      {self.key} ({arg_list}) - {self.docstring}\nUSAGE:\n      {self.usage}"
        return f"      {self.key} ({arg_list}) - {self.docstring}"



import os
import sys
from typing import Callable

from prompt_toolkit import print_formatted_text, prompt
from prompt_toolkit.shortcuts.prompt import confirm


class CmdShortCuts:
    @property
    def prompt(self) -> Callable:
        return prompt

    @property
    def print_ft(self) -> Callable:
        return print_formatted_text

    @property
    def confirm(self) -> Callable:
        return confirm

    @staticmethod
    def clear() -> None:
        os.system("cls") if sys.platform == "win32" else os.system("clear")



import inspect
import sys
from typing import Any, Type, Union, get_args, get_origin

NoneType = type(None)


class TypeCaster:
    @classmethod
    def _typing_to_builtin(cls, type_hint: Type) -> Type:
        origin = get_origin(type_hint)
        args = get_args(type_hint)

        if origin is not None and args:
            # Recursively convert the nested generic types
            converted_args = tuple(cls._typing_to_builtin(arg) for arg in args)
            return origin[converted_args]
        else:
            return type_hint

    @classmethod
    def cast(cls, type_hint: Type, value: Any) -> Any:
        # extracted annotation from `inspect.get_annotations`
        if type_hint is inspect.Parameter.empty:
            return value

        if sys.version_info >= (3, 9):
            type_hint = cls._typing_to_builtin(type_hint)

        origin = get_origin(type_hint)
        args = get_args(type_hint)
        # None
        if value is None and type_hint is not bool:
            return value

        if origin is not None and args:
            # list
            if origin is list:
                return [cls.cast(type_hint=args[0], value=v) for v in value]
            # dict
            elif origin is dict:
                key_type, value_type = args
                return {
                    cls.cast(type_hint=key_type, value=k): cls.cast(type_hint=value_type, value=v)
                    for k, v in value.items()
                }
            # Optional
            elif origin is Union:
                if value is None and NoneType in args:
                    return None
                # in python3.8 raise TypeError: issubclass() arg 1 must be a class
                # example _cast_type(Optional[List[int]], [])
                non_none_args = [arg for arg in args if arg is not NoneType]
                if len(non_none_args) == 1:
                    return cls.cast(type_hint=non_none_args[0], value=value)
        # bool cast
        elif type_hint is bool:
            return bool(value)
        else:
            # direct cast
            return type_hint(value)
